vec4 overviewTableRowBg = vec4(.2, .2, .2, .2);
vec4 ColorYellow = vec4(.8, .6, .1, 1);
vec4 ColorLightGrey = vec4(.7, .7, .7, .6);
vec4 ColorExtraLightGrey = vec4(.8, .8, .8, .7);
vec4 ColorGrey = vec4(.5, .5, .5, .5);
vec4 ColorBlue = vec4(0, 0, 1, 1);
vec4 ColorPurple = vec4(.5, 0, .5, 1);
vec4 ColorGreen = vec4(0, .6, 0, 1);
vec4 ColorRed = vec4(.8, 0, 0, 1);
UI::Font@ MidFont = UI::LoadFont("DroidSans.ttf", 20, -1, -1, true, true, true);
UI::Font@ BoldFont = UI::LoadFont("DroidSans-Bold.ttf", 16, -1, -1, true, true, true);
UI::Font@ BigFont = UI::LoadFont("DroidSans.ttf", 26, -1, -1, true, true, true);

enum menuTabs {
    None,
    Campaigns,
    mapsByCampaign,
    About,
    News,
    activePool,
    map,
    debug,
    tag,
    tagInfo
}

class stringMenuTabsPair {
    string name;
    menuTabs tab;

    stringMenuTabsPair(string _name, menuTabs _tab) {
        name = _name;
        tab = _tab;
    }
}

stringMenuTabsPair@[] menuTabFromStringArray = {stringMenuTabsPair("None", menuTabs::None),
    stringMenuTabsPair("Campaigns", menuTabs::Campaigns),
    stringMenuTabsPair("mapsByCampaign", menuTabs::mapsByCampaign),
    stringMenuTabsPair("About", menuTabs::About),
    stringMenuTabsPair("News", menuTabs::News),
    stringMenuTabsPair("activePool", menuTabs::activePool),
    stringMenuTabsPair("map", menuTabs::map),
    stringMenuTabsPair("debug", menuTabs::debug),
    stringMenuTabsPair("tag", menuTabs::tag),
    stringMenuTabsPair("tagInfo", menuTabs::tagInfo)
};

menuTabs menuTabFromString(const string &in tab) {
    for (uint i = 0; i < menuTabFromStringArray.Length; i++) {
        if (tab == menuTabFromStringArray[i].name) return menuTabFromStringArray[i].tab;
    }
    return menuTabs::None;
}

void renderInnerWindow() {
    menuBar();
}

menuTabs selectedTab = menuTabs::News;
string newsSelctedTab = "None";

bool menuTab(const string &in name, menuTabs tab) {
    int flags = selectedTab == tab ? UI::TabItemFlags::SetSelected : UI::TabItemFlags::None;
    return UI::BeginTabItem(name, flags);
}

bool menuTab(const string &in name, menuTabs tab, bool&out open) {
    int flags = selectedTab == tab ? UI::TabItemFlags::SetSelected : UI::TabItemFlags::None;
    return UI::BeginTabItem(name, open, flags);
}

bool newsTab(const string &in name) {
    int flags = (newsSelctedTab == name ? UI::TabItemFlags::SetSelected : UI::TabItemFlags::None) | UI::TabItemFlags::NoPushId;
    return UI::BeginTabItem(name, flags);
}

void menuBar() {
    UI::BeginTabBar("menu bar");
    
    if (menuTab("News", menuTabs::News)) {
        drawNews();
        UI::EndTabItem();
    }

    if (menuTab("Active Map Pool", menuTabs::activePool)) {
        drawActiveMapPool();
        UI::EndTabItem();
    }

    if (menuTab("Campaigns", menuTabs::Campaigns)) {
        drawAllCampaigns();
        UI::EndTabItem();
    }

    if (menuTab("Maps", menuTabs::mapsByCampaign)) {
        drawMaps();
        UI::EndTabItem();
    }

    if (menuTab("Tags", menuTabs::tag)) {
        tag::GUI::drawTab();
        UI::EndTabItem();
    }

    if (menuTab("Settings", menuTabs::About)) {
        drawAbout();
        UI::EndTabItem();
    }

    UI::PushStyleColor(UI::Col::TabActive, ColorYellow);
    UI::PushStyleColor(UI::Col::TabHovered, ColorYellow * vec4(1.2f, 1.2f, 1.2f, 0.85f));
    UI::PushStyleColor(UI::Col::Tab, ColorYellow * vec4(0.5f, 0.5f, 0.5f, 0.75f));
    // for temporary tabs
    if (isMapSelected) {
        bool open;
        if (menuTab("Map info", menuTabs::map, open)) {
            drawMapInfoTabInner();
            UI::EndTabItem();
        }
        if (!open) {
            isMapSelected = false;
        }
    }

    if (tag::showTagEdit) {
        bool open;
        if (menuTab("Edit tag", menuTabs::tagInfo, open)) {
            tag::currentTagEdit.drawEditTabInner();
            UI::EndTabItem();
        }
        if (!open) tag::showTagEdit = false;
    }

    if (currentDebugType != debugType::none) {
        bool open;
        if (develepor && menuTab("Debug", menuTabs::debug, open)) {
            drawDebug();
            UI::EndTabItem();
        }
        if (!open) currentDebugType = debugType::none;
    }

    UI::PopStyleColor(3);

    UI::EndTabBar();
    selectedTab = menuTabs::None;
}

void drawAbout() {
    UI::PushFont(MidFont);
    if (UI::Button("Hard Reload Memory")) {
        startnew(CoroutineFunc(hardReload));
    }
    UI::PopFont();
}

void hardReload() {
    memory.Clear();
    loadMaps();
}

bool newsInitialized = false;

newsItem@[] newsItems = {};

void initiateNews() {
    Net::HttpRequest@ req = Net::HttpGet(develepor ? "https://openplanet.dev/plugin/royal-hub/config/news-dev" : "https://openplanet.dev/plugin/royal-hub/config/news");
    while (!req.Finished()) yield();
    if (req.ResponseCode() != 200) {
        error("bad response code. Check network log!");
        trace(develepor ? "https://openplanet.dev/plugin/royal-hub/config/news-dev" : "https://openplanet.dev/plugin/royal-hub/config/news");
        return;
    }

    Json::Value json = req.Json();

    for (uint i = 0; i < json.Length; i++) {
        if (!isOlderVersion(json[i]["versionRequirement"])) continue;
        if (json[i].HasKey("disabled") && bool(json[i]["disabled"])) continue;
        newsItems.InsertLast(newsItem(json[i]));
    }

    startnew(newsTabHandler);
    newsInitialized = true;
}

bool isOlderVersion(const string &in version) {
    string[] versionArr = version.Split('.');
    for (int i = 0; i < Math::Min(versionArr.Length, pluginVersionArr.Length); i++) {
        if (Text::ParseUInt(versionArr[i]) > pluginVersionArr[i]) return false;
        if (Text::ParseUInt(versionArr[i]) < pluginVersionArr[i]) return true;
    }

    return versionArr.Length <= pluginVersionArr.Length;
}

string icons(const string &in inp) {
    if (inp == "YoutubePlay") return Icons::YoutubePlay;
    if (inp == "DiscordAlt") return Icons::DiscordAlt;
    if (inp == "Google") return Icons::Google;
    if (inp == "Play") return Icons::Play;
    return "";
}


void drawNews() {
    if (!newsInitialized) {
        UI::Text("Loading");
        return;
    }
    UI::PushStyleColor(UI::Col::TabActive, ColorLightGrey);
    UI::PushStyleColor(UI::Col::TabHovered, ColorExtraLightGrey);
    UI::PushStyleColor(UI::Col::TabUnfocused, ColorGrey);
    UI::PushStyleColor(UI::Col::Tab, ColorGrey);
    UI::BeginTabBar("News");

    for (uint i = 0; i < newsItems.Length; i++) {
        if (newsTab(newsItems[i].title)) {
            currentNewsTab = i;
            newsItems[i].draw();
            UI::EndTabItem();
        }
    }

    UI::EndTabBar();
    UI::PopStyleColor(4);

    newsSelctedTab = "None";
    newsScroll = UI::GetScrollY();
}

enum imgScaleType {
    scaleToMax,
    scaleToMin,
    stretch,
    scaleToWidth,
    scaleToHeight,
    noScale
}

vec2 scaleImage(vec2 img, float offset, imgScaleType scaleType = imgScaleType::scaleToWidth, vec2 screen = UI::GetWindowSize()) {
    screen.y -= offset;
    float r_width = screen.x / img.x;
    float r_height = screen.y / img.y;

    if (scaleType == imgScaleType::scaleToMax) return img * Math::Max(r_width, r_height);
    if (scaleType == imgScaleType::scaleToMin) return img * Math::Min(r_width, r_height);
    if (scaleType == imgScaleType::stretch) {
        img.x *= r_width;
        img.y *= r_height;
        return img;
    }
    if (scaleType == imgScaleType::scaleToWidth) return img * r_width;
    if (scaleType == imgScaleType::scaleToHeight) return img * r_height;
    return img;

}


uint currentNewsTab = 0;
uint lastNewsTab = 0;
uint newsTimeLastSwitch = 0;
uint newsTimeBetweenSwitch = 10000;
float newsScroll = 0;

void newsTabHandler() {
    newsTimeLastSwitch = Time::Now;
    while (true) {
        if (lastNewsTab != currentNewsTab) {
            newsTimeLastSwitch = Time::Now;
            lastNewsTab = currentNewsTab;
        }
        if (newsScroll != 0) {
            newsTimeLastSwitch = Math::Max(newsTimeLastSwitch, Time::Now - newsTimeBetweenSwitch/4);
        }
        if (Time::Now - newsTimeLastSwitch >= newsTimeBetweenSwitch) {
            newsNextTab();
        }
        yield();
    }
}

void newsNextTab() {
    currentNewsTab = (currentNewsTab + 1) % newsItems.Length;
    newsSelctedTab = newsItems[currentNewsTab].title;
}

void drawAllCampaigns() {
    if (campaignsLoading > 0) UI::ProgressBar(1 - float(campaignsLoading)/campaignAmmount);
    if (campaignsLoading > 0) UI::Separator();
    UI::PushFont(MidFont);
    drawGlobalMedals();
    UI::PopFont();
    UI::Separator();
    if (UI::BeginChild("all-campaigns")) {
        UI::PushStyleColor(UI::Col::TableRowBg, overviewTableRowBg);
        if(UI::BeginTable("campaigns-table", 3, UI::TableFlags::SizingFixedFit | UI::TableFlags::RowBg)) {
            UI::TableSetupColumn("name", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("This is how long a map", MidFont).x);
            UI::TableSetupColumn("medals", UI::TableColumnFlags::WidthStretch);
            UI::TableSetupColumn("buttons");
            drawAllCampaignsTable();
            UI::EndTable();
        }
        UI::PopStyleColor();
    }
    UI::EndChild();
}

void drawGlobalMedals() {
    uint[] gMedals = array<uint>(6, 0);
    uint gMax = 0;
    for (uint i = 0; i < orderedKeys.Length; i++) {
        campaignInfo@ ci = memory.Get(orderedKeys[i]);
        uint[] iMedals = ci.countMedals();
        for (uint j = 0; j < gMedals.Length; j++) {
            gMedals[j] += iMedals[j];
        }
        gMax += ci.maps.Length;
    }

    UI::Text(getMedalIcon(0) + gMedals[0] + "/" + gMax);
    UI::SameLine();
    UI::Text(getMedalIcon(1) + gMedals[1] + "/" + gMax);
    UI::SameLine();
    UI::Text(getMedalIcon(2) + gMedals[2] + "/" + gMax);
    UI::SameLine();
    UI::Text(getMedalIcon(3) + gMedals[3] + "/" + gMax);
    UI::SameLine();
    UI::Text(getMedalIcon(4) + gMedals[4] + "/" + gMax);
    UI::SameLine();
    UI::Text(getMedalIcon(5) + gMedals[5] + "/" + gMax);
}

void drawAllCampaignsTable() {
    while (!memory.initialized) return;
    string[] keys = orderedKeys;
    UI::ListClipper clip(keys.Length);
    while (clip.Step()) {
        for (int i = clip.DisplayStart; i < clip.DisplayEnd; i++) {
            memory.Get(keys[i]).DrawTableRow();
        }
    }
}

void drawMaps() {
    if (campaignsLoading > 0) UI::ProgressBar(1 - float(campaignsLoading)/campaignAmmount);
    if (campaignsLoading > 0) UI::Separator();
    UI::PushFont(MidFont);
    if (mapsFilter.drawFilter()) {
        filterMaps();
    }
    drawGlobalMedals();
    drawRandomiserBlock();
    UI::PopFont();
    UI::Separator();
    drawFilteredMapsTable();
}

mapInfo@[] filteredMaps;

void filterMaps(mapInfo@[]@ outList = filteredMaps, mapFilter@ filter = mapsFilter) {
    outList = {};
    for (uint i = 0; i < orderedKeys.Length; i++) {
        memory.Get(orderedKeys[i]).getFilterMaps(outList, filter);
    }
}

void drawFilteredMapsTable(mapInfo@[] mapsToShow = filteredMaps) {
    while (!memory.initialized) return;
    if (UI::BeginTable("campaign-maps-table", 4, UI::TableFlags::SizingFixedFit | UI::TableFlags::RowBg)) {
        UI::TableSetupColumn("name", UI::TableColumnFlags::WidthStretch);
        UI::TableSetupColumn("medal-icon", UI::TableColumnFlags::WidthFixed, Draw::MeasureString(getMedalIcon(0)).x+5);
        UI::TableSetupColumn("time", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("0:00.000").x+5);
        UI::TableSetupColumn("buttons");

        
        UI::ListClipper clip(mapsToShow.Length);
        while (clip.Step()) {
            for (int i = clip.DisplayStart; i < clip.DisplayEnd; i++) {
                mapsToShow[i].drawMap();
            }
        }

        UI::EndTable();
    }
}   

void drawRandomiserBlock() {
    if (UI::Button(Icons::Play+"Random")) {
        startnew(CoroutineFunc(playRandomFilterMap));
    }
    UI::SameLine();
    UI::Text(tostring(filteredMaps.Length) + " maps");
}

void playRandomFilterMap() {
    if (filteredMaps.Length == 0) return;
    filteredMaps[Math::Rand(0, filteredMaps.Length)].playMap();
}

void resetFilters() {
    mapsFilter = mapFilter();
    filterMaps();
}

void drawActiveMapPool() {
    drawActiveMapPoolTable();
}

void drawActiveMapPoolTable() {
    float width = UI::GetWindowContentRegionWidth();
    double colWidth = width / 6.;
    auto fp = UI::GetStyleVarVec2(UI::StyleVar::FramePadding);
    vec2 btnSize = vec2(colWidth - fp.x, 100.);
    UI::Dummy(vec2(colWidth/2. - fp.x, 0));
    UI::SameLine();
    if (UI::BeginTable("active-map-pool-table", 5)) {
        for (uint i = 0; i < 5; i++) {
            UI::TableSetupColumn(tostring(i), UI::TableColumnFlags::WidthFixed, btnSize.x);
        }

        drawActiveMapPoolTableInner(btnSize);
        UI::EndTable();
    }
}

void drawActiveMapPoolTableInner(vec2 btnSize) {
    for (int i = 0; i < Math::Min(activePoolAmmount, orderedKeys.Length); i++) {
        //if (i%5 == 0) UI::TableNextRow();
        UI::TableNextColumn();
        memory.Get(orderedKeys[i]).drawCampaignButton(btnSize);
    }
}

void drawActiveMapPoolCampaign(campaignInfo@ ci, vec2 btnSize) {
    if (UI::ButtonColored(ci.cleanName, .1, .9, .9, btnSize)) {
        ci.playCampaign();
    }
}


mapInfo@ currentSelectedMap = mapInfo();
bool isMapSelected = false;
uint mxId = 0;
UI::Texture@ mapInfoThumbnail;
string mxUrl = "trackmania.exchange";

void setMapInfo(mapInfo@ mi) {
    @currentSelectedMap = mi;
    @mapInfoThumbnail = null;
    isMapSelected = true;
    selectedTab = menuTabs::map;
    mxId = 0;

    // load image
    Net::HttpRequest@ req;
    @req = Net::HttpGet(mi.thumbnailUrl);

    while(!req.Finished()) yield();

    if (req.ResponseCode() != 200) {
        error("bad response code. Check network log!");
        return;
    }

    @mapInfoThumbnail = UI::LoadTexture(req.Buffer());

    mxId = TMXAPIMapIdReq(mi.trackUid);

    debug("mxId: "+ mxId);
}

void drawMapInfoTabInner() {
    if (!isMapSelected) {
        UI::Text("Please Select a map");
        return;
    }
    drawMapInfoThumbnail(); // initilises a child "imageBar" remeber to call EndChild
    
    UI::Separator();
    drawMapInfoInfo();
    UI::Separator();

    UI::PushStyleColor(UI::Col::Button, ColorGreen);
    UI::PushStyleColor(UI::Col::ButtonHovered, ColorGreen*1.1);
    UI::PushStyleColor(UI::Col::ButtonActive, ColorGreen*0.9);

    if (UI::Button(Icons::Play + " Play this map")) {
        currentSelectedMap.playMap();
    }
    
    if (mxId != 0 && UI::Button(Icons::ExternalLink + " View on Trackmania exchange")) {
        OpenBrowserURL("https://"+mxUrl+"/maps/"+mxId);
    }

    UI::PopStyleColor(3);

    UI::PushStyleColor(UI::Col::Button, ColorPurple);
    UI::PushStyleColor(UI::Col::ButtonHovered, ColorPurple*1.1);
    UI::PushStyleColor(UI::Col::ButtonActive, ColorPurple*0.9);
    
    if (develepor && UI::Button(Icons::Search + " Debug this map")) {
        startnew(CoroutineFunc(currentSelectedMap.debugThisMap));
    }

    UI::PopStyleColor(3);

    UI::EndChild();
    UI::SameLine();
    UI::BeginChild("titleSide");

    UI::PushFont(BigFont);
    UI::Text(currentSelectedMap.name);
    UI::PopFont();

    drawMapInfoMedals();
    UI::SameLine();
    drawMapTagInfo();

    drawExtraMapInfoTabs();

    UI::EndChild();
}

void drawMapTagInfo() {
    UI::BeginChild("tagsInfo", vec2(1000, 100));
    if (!tagMemory.initialized) {
        UI::Text("loading tags");
        return;
    }

    uint64[] currentMapTags = currentSelectedMap.tags;
    for (uint i = 0; i < currentMapTags.Length; i++) {
        if (!tagMemory.In("" + currentMapTags[i])) continue;
        UI::SameLine();
        tagMemory.Get("" + currentMapTags[i]).drawTagButton();
    }
    UI::EndChild();
}

void drawExtraMapInfoTabs() {
    UI::BeginChild("extraMapInfoTabs");
    UI::BeginTabBar("extraMapInfoMenuTabs");

    if (UI::BeginTabItem("Tags")) {
        tag::GUI::drawTagInfo(true, currentSelectedMap);
        UI::EndTabItem();
    }

    /* coming soon
    if (UI::BeginTabItem("Leader board")) {

        UI::EndTabItem();
    }//*/

    UI::EndTabBar();
    UI::EndChild();
}

void drawMapInfoInfo() {
    UI::Text(getDifficultyIcon(currentSelectedMap.getDifficulty()) + " \\$f77" + getDifficultyName(currentSelectedMap.getDifficulty()));
    UI::SetItemTooltip("Difficulty");

    if (currentSelectedMap.getOld()) {
        UI::Text(Icons::Times + " \\$f77" + "Updated version exist");
    } else {
        UI::Text(Icons::Check + " \\$f77" + "Is newest version");
    }
    UI::SetItemTooltip("Version");
}

void drawMapInfoMedals() {
    UI::BeginChild("medals", vec2(155, currentSelectedMap.time != 0 ? 120 : 100));
    if (UI::BeginTable("medals", 3)) {
        UI::TableSetupColumn("", UI::TableColumnFlags::WidthFixed, Draw::MeasureString(getMedalIcon(0)).x);
        UI::TableSetupColumn("Time", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("0:00.000").x);
        UI::TableSetupColumn("Delta", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("+0:00.000").x);
        UI::TableHeadersRow();

        int playerTime = currentSelectedMap.time;
        bool showedPlayer = (playerTime == 0);

        for (uint i = 0; i < currentSelectedMap.medals.Length; i++) {
            if (!showedPlayer && playerTime < currentSelectedMap.medals[i]) {
                showedPlayer = true;
                UI::TableNextRow();
                UI::TableNextColumn();
                UI::Text(getMedalIcon(5-i));
                UI::TableNextColumn();
                UI::Text(Time::Format(playerTime));
            }
            UI::TableNextRow();
            UI::TableNextColumn();
            UI::Text(getMedalIcon(5-i));
            UI::TableNextColumn();
            UI::Text(Time::Format(currentSelectedMap.medals[i]));
            if (playerTime != 0) {
                UI::TableNextColumn();
                string leadout = "\\$3C3";
                if (currentSelectedMap.medals[i]-playerTime < 0) leadout = "\\$F00+";
                UI::Text(leadout + Time::Format(playerTime - currentSelectedMap.medals[i]) + "\\$z");
            }
        }

        if (!showedPlayer) {
                showedPlayer = true;
                UI::TableNextRow();
                UI::TableNextColumn();
                UI::Text(getMedalIcon(1));
                UI::TableNextColumn();
                UI::Text(Time::Format(playerTime));
            }

        UI::EndTable();
    }
    UI::EndChild();

}

void drawMapInfoThumbnail() {
    float imageBarWidth = UI::GetWindowSize().x*0.35;

    UI::BeginChild("imageBar", vec2(imageBarWidth, 0));

    if (mapInfoThumbnail is null) {
        int HourGlassValue = Time::Stamp % 3;
        string Hourglass = (HourGlassValue == 0 ? Icons::HourglassStart : (HourGlassValue == 1 ? Icons::HourglassHalf : Icons::HourglassEnd));
        UI::Text(Hourglass + " Loading");
    } else {
        UI::Image(mapInfoThumbnail, scaleImage(mapInfoThumbnail.GetSize(), 0, imgScaleType::scaleToWidth, vec2(imageBarWidth, UI::GetWindowSize().y)));
    }
}