class mapFilter {
    string name = "";
    bool nameOn = false;
    bool[] medals = array<bool>(6, true); // unplayed, played, bronze, silver, gold, author
    bool[] difficulty = array<bool>(6, true); // white, green, blue, red, black, other
    bool old = true; // old maps witch has gotten updates
    uint64[] tagInclusive = {};
    uint64[] tagExclusive = {};
    bool includeAllTags = false; // maps need to include all tags from tagInclusive

    bool lnameOn = false;
    bool[] lmedals = array<bool>(6, true); // unplayed, played, bronze, silver, gold, author
    bool[] ldifficulty = array<bool>(6, true); // white, green, blue, red, black, other
    bool lold = true; // old maps witch has gotten updates
    bool lincludeAllTags = false; // maps need to include all tags from tagInclusive

    mapFilter() {}

    bool filterMap(mapInfo@ mi) {
        if (!medals[mi.getMedal()]) return false;
        if (!difficulty[mi.getDifficulty()]) return false;

        if (nameOn) {
            string str1 = name.ToLower(), str2 = mi.name.ToLower();
            if (!str2.Contains(str1)) return false;
        }

        if (!old && mi.getOld()) return false;

        if (!mi.hasTagList(tagInclusive, includeAllTags)) return false;
        if (tagExclusive.Length > 0 && mi.hasTagList(tagExclusive, false)) return false;

        return true;
    }

    bool drawFilter() {
        bool changed = false;
        name = UI::InputText("Name", name, changed);

        // tags
        changed = changed || drawTagList(true);
        changed = changed || drawTagList(false);
        

        old = UI::Checkbox("Include outdated maps", old);
        changed = changed || lold != old;
        lold = old;

        for (uint i = 0; i < medals.Length; i++) {
            medals[i] = UI::Checkbox(getMedalIcon(i), medals[i]);
            changed = changed || lmedals[i] != medals[i];
            lmedals[i] = medals[i];
            UI::SameLine();
        }

        UI::NewLine();

        for (uint i = 0; i < difficulty.Length; i++) {
            difficulty[i] = UI::Checkbox(getDifficultyIcon(i), difficulty[i]);
            changed = changed || ldifficulty[i] != difficulty[i];
            ldifficulty[i] = difficulty[i];
            UI::SameLine();
        }
        
        if (UI::Button(Icons::Refresh)) {
            filterMaps();
        }

        UI::SameLine();
        if (UI::Button(Icons::Times)) {
            resetFilters();
        }

        
        
        nameOn = name != "";

        return changed;
    }

    bool drawTagList(bool include) {
        bool changed = false;
        uint64[]@ list = include ? tagInclusive : tagExclusive;
        if (UI::CollapsingHeader(include ? "Include tags" : "Exclude tags")) {
            uint spaceUsedOnLine = 0;

            if (true) {
                UI::PushID("allTags");
                bool isSelected = list.Length == allTagUids.Length;
                bool lisSelected = isSelected;
                UI::PushStyleColor(UI::Col::ChildBg, ColorGrey);
                UI::PushStyleColor(UI::Col::CheckMark, ColorGrey);
                UI::BeginChild(include ? "tagIncAllTags" : "tagExcAllTags", Draw::MeasureString("All tags", MidFont) + vec2(40, 8));
                isSelected = UI::Checkbox("All tags", isSelected);
                UI::EndChild();
                UI::PopStyleColor(2);
                if (isSelected != lisSelected) {
                    changed = true;
                    if (isSelected) {
                        list = allTagUids;
                    } else {
                        list = {};
                    }
                }
                UI::PopID();
                if (include) {
                    UI::SameLine();
                    UI::PushID("includeAllTags");
                    UI::PushStyleColor(UI::Col::ChildBg, ColorGrey);
                    UI::PushStyleColor(UI::Col::CheckMark, ColorGrey);
                    UI::BeginChild("includeAllTagsChild", Draw::MeasureString("Include all selected tags", MidFont) + vec2(40, 8));
                    includeAllTags = UI::Checkbox("Include all selected tags", includeAllTags);
                    UI::EndChild();
                    UI::PopStyleColor(2);
                    changed = changed || includeAllTags != lincludeAllTags;
                    lincludeAllTags = includeAllTags;
                    UI::PopID();
                }
            }

            for (uint i = 0; i < allTagUids.Length; i++) {
                tag@ curTag = tagMemory.Get(""+ allTagUids[i]);
                UI::PushID((include ? "tagInc" : "tagExc") +i);
                bool isSelected = list.Find(allTagUids[i]) != -1;
                bool lisSelected = isSelected;
                UI::PushStyleColor(UI::Col::ChildBg, curTag.colorVec4);
                UI::PushStyleColor(UI::Col::CheckMark, curTag.colorVec4);
                spaceUsedOnLine += Draw::MeasureString(curTag.name, MidFont).x + 40 + 10;
                if (spaceUsedOnLine > UI::GetWindowSize().x) {
                    spaceUsedOnLine = Draw::MeasureString(curTag.name, MidFont).x + 40 + 10;
                    UI::NewLine();
                }
                UI::BeginChild(include ? "tagIncChild"+i : "tagExcChild"+i, Draw::MeasureString(curTag.name, MidFont) + vec2(40, 8));
                isSelected = UI::Checkbox(curTag.name, isSelected);
                UI::EndChild();
                UI::PopStyleColor(2);
                UI::SameLine();
                if (isSelected != lisSelected) {
                    if (isSelected) {
                        list.InsertLast(allTagUids[i]);
                        changed = true;
                    } else {
                        list.RemoveAt(list.Find(allTagUids[i]));
                        changed = true;
                    }
                }
                UI::PopID();
            }
            UI::NewLine();
            if (changed) {
                debugRef(include ? "tagInclusive" : "tagExclusive", list);
            }
        }
        return changed;
    }
}