class mapInfo {
    string trackId = "";
    string trackUid = "";
    string campaignId = "";
    string name = "";
    string author = "";
    string fileUrl = "";
    string thumbnailUrl = "";
    uint64[] medals = {0, 0, 0, 0};
    uint64 time = 0;
    uint difficulty = 5; // white, green, blue, red, black, other
    bool isOld = false; 
    uint64[] tags = {};

    mapInfo() {}

    mapInfo(string _trackId, string _trackUid, string _campaignId, string _name, string _author, string _fileUrl, uint64[] _medals, string _thumbnailUrl, uint64 _time = 0, bool _isOld = false, uint64[] _tags = {}) {
        trackId = _trackId;
        trackUid = _trackUid;
        campaignId = _campaignId;
        name = _name;
        author = _author;
        fileUrl = _fileUrl;
        thumbnailUrl = _thumbnailUrl;
        medals = _medals;
        time = _time;
        difficulty = difficultyByName(name);
        isOld = _isOld;
        tags = _tags;

        if (overide::mapOverideDict.Exists(trackId)) {
            overide(cast<mapInfo::mapInfoOveride@>(overide::mapOverideDict[trackId]));
        }
    }

    void overide(mapInfo::mapInfoOveride@ mio) {
        if (mio.difficulty != -1) difficulty = mio.difficulty;
        if (mio.isOld != -1) isOld = (mio.isOld != 0);
        if (mio.name != "") name = mio.name;
    }

    void writeToBuffer(MemoryBuffer@ buf) {
        stringToBuffer(buf, trackId);
        stringToBuffer(buf, trackUid);
        stringToBuffer(buf, campaignId);
        stringToBuffer(buf, name);
        stringToBuffer(buf, author);
        stringToBuffer(buf, fileUrl);
        stringToBuffer(buf, thumbnailUrl);
        uint64ArrayToBuffer(buf, medals);
        buf.Write(time);
        boolToBuffer(buf, isOld);
    }

    uint getMedal() {
        if (time == 0) return 0;
        for (int i = medals.Length-1; i > -1; i--) {
            if (time <= medals[i]) continue;
            return medals.Length-i;
        }
        return 5;
    }

    uint getDifficulty() {
        return difficulty;
    }

    bool getOld() {
        return isOld;
    }

    void setOld(bool old) {
        isOld = old;
    }

    void drawMap() {
        UI::PushID(trackId);
        UI::TableNextRow();
        UI::TableNextColumn();
        UI::Text(name);
        UI::TableNextColumn();
        UI::Text(getMedalIcon(getMedal()));
        UI::TableNextColumn();
        if (time != 0) {
            UI::Text(Time::Format(time));
        }
        UI::TableNextColumn();
        if (UI::Button(Icons::Play)) {
            startnew(CoroutineFunc(playMap));
        }

        UI::SameLine();

        if (UI::Button(Icons::Search)) {
            startnew(CoroutineFunc(viewThisMap));
        }

        if (develepor) {
            UI::SameLine();
            if (UI::Button("\\$F6F" + Icons::Search + "\\$z")) {
                startnew(CoroutineFunc(debugThisMap));
            }
        }

        UI::PopID();
    }

    void debugThisMap() {
        debugMap(this);
    }

    void viewThisMap() {
        setMapInfo(this);
        debugRef("tags", tags);
    }

    void playMap() {
        if (!Permissions::PlayLocalMap()) {
            error("User cant play local maps!");
            return;
        }
        auto app = cast<CGameManiaPlanet>(GetApp());
        if (app.Network.PlaygroundClientScriptAPI.IsInGameMenuDisplayed) {
            app.Network.PlaygroundInterfaceScriptHandler.CloseInGameMenu(CGameScriptHandlerPlaygroundInterface::EInGameMenuResult::Quit);
        }
        app.BackToMainMenu();
        while (!app.ManiaTitleControlScriptAPI.IsReady) yield();
        app.ManiaTitleControlScriptAPI.PlayMap(fileUrl, "", "");
        mapWatch.setMap(this);
        UI::HideOverlay();
        viewThisMap();
    }

    void drawMapDebug() {
        UI::Text("Debbuging map");
        UI::Text("trackId : " + trackId);
        UI::SameLine();
        if (UI::Button(Icons::Clipboard)) {
            IO::SetClipboard(trackId);
        }
        UI::Text("trackUid : " + trackUid);
        UI::Text("campaignId : " + campaignId);
        UI::Text("name : " + name);
        UI::Text("author : " + author);
        UI::Text("fileUrl : " + fileUrl);
        UI::Text("thumbnailUrl : " + thumbnailUrl);
        UI::Text("medals : [" + medals[0] + ", " + medals[1] + ", " + medals[2] + ", " + medals[3] + "]");
        UI::Text("time : " + time);
        UI::Text("difficulty : " + difficulty);
        UI::Text("isOld : " + isOld);

        UI::Separator();

        UI::Text("getMedal() : " + getMedal());
        UI::Text("getDifficulty() : " + getDifficulty());
    }

    void drawQuickDebug() {
        if (UI::Button(Icons::Search)) {
            startnew(CoroutineFunc(debugThisMap));
        }
        UI::SameLine();
        UI::Text("name : " + name + ", trackId : " + trackId + ", trackUid : " + trackUid);
    }

    void loadTag(uint tagUid) {
        if (hasTag(tagUid)) return;
        tags.InsertLast(tagUid);
        tagMemory.Get(""+tagUid).mapsWithTag++;
    }

    void loadAllTags() {
        tag::campaignMapId@ tagMapId = tag::campaignMapId(campaignId, trackId);
        while (!tagMemory.initialized) yield();
        for (uint i = 0; i < allTagUids.Length; i++) {
            if (tagMemory.Get(""+allTagUids[i]).maps.Find(tagMapId) != -1) {
                tags.InsertLast(allTagUids[i]);
            }
            yield();
        }
    }

    void addTag(uint tagUid) {
        tags.InsertLast(tagUid);
        tagMemory.Get(""+tagUid).addMap(this);
        tagMemory.Set(""+tagUid, tagMemory.Get(""+tagUid));
        debugRef("tags", tags);
    }

    void removeTag(uint tagUid) {
        tags.RemoveAt(tags.Find(tagUid));
        tagMemory.Get(""+tagUid).removeMap(this);
        tagMemory.Set(""+tagUid, tagMemory.Get(""+tagUid));
    }

    bool hasTag(uint tagUid) {
        return tags.Find(tagUid) != -1;
    }

    bool hasTagList(uint64[] tagUids, bool includeAll = false) {
        if (tagUids.Length == 0) return true;
        if (tags.Length < tagUids.Length && includeAll) return false;
        tags.SortAsc();
        tagUids.SortAsc();

        uint a = 0, b = 0;
        while (a < tags.Length && b < tagUids.Length) {
            if (tags[a] == tagUids[b]) {
                if (!includeAll) return true;
                a++; b++; continue;
            }
            if (tags[a] < tagUids[b]) {
                a++; continue;
            }
            b++;
            if (includeAll) return false;
        }
        return (b == tagUids.Length && includeAll);
    }
}

namespace mapInfo {
    mapInfo@ readFromBuffer(MemoryBuffer@ buf) {
        string trackId = StringFromBuffer(buf);
        string trackUid = StringFromBuffer(buf);
        string campaignId = StringFromBuffer(buf);
        string name = StringFromBuffer(buf);
        string author = StringFromBuffer(buf);
        string fileUrl = StringFromBuffer(buf);
        string thumbnailUrl = StringFromBuffer(buf);
        uint64[] medals = uint64ArrayFromBuffer(buf);
        uint64 time = uint64FromBuffer(buf);
        bool isOld = boolFromBuffer(buf);
        mapInfo@ mi = mapInfo(trackId, trackUid, campaignId, name, author, fileUrl, medals, thumbnailUrl, time, isOld);
        return mi;
    }

    class mapInfoOveride {
        int difficulty = -1; // white, green, blue, red, black, other
        int isOld = -1; 
        string name = "";

        mapInfoOveride() {}

        mapInfoOveride(int _difficulty, int _isOld, string _name) {
            difficulty = _difficulty;
            isOld = _isOld;
            name = _name;
        }
    }
}