class newsItem {
    string title;
    bool working;
    UI::Texture@ Image;
    Json::Value Buttons;
    Json::Value Description;

    newsItem() {
        title = "";
        working = false;
        @Image = null;
        Buttons = Json::Object();
        Description = Json::Object();
    }

    newsItem(string _title, UI::Texture@ _image, Json::Value _Buttons, Json::Value _Description) {
        title = _title;
        @Image = _image;
        Buttons = _Buttons;
        Description = _Description;
        working = true;
    }

    newsItem(Json::Value json) {
        if (!isOlderVersion(json["versionRequirement"])) {
            error("Tried to make newsitem from newer version, try updating plugin");
            return;
        }

        if (json.HasKey("disabled") && bool(json["disabled"])) return;

        title = json["name"];
        Net::HttpRequest@ req;
        if (json.HasKey("imageUrl")) @req = Net::HttpGet(json["imageUrl"]);
        else if (json.HasKey("campaignImage")) {
            Json::Value json2 = nadeoAPIReq("https://live-services.trackmania.nadeo.live/api/token/club/"+clubId+"/activity?length=1&offset="+ int(json["campaignImage"]) +"&active=true");
            @req = Net::HttpGet(json2["activityList"][0]["mediaUrlPngLarge"]);
        }
        else {
            error("No image found in news, try contacting Gr1sen");
            return;
        }

        while(!req.Finished()) yield();
        if (req.ResponseCode() != 200) {
            error("bad response code. Check network log!");
            return;
        }

        @Image = UI::LoadTexture(req.Buffer());

        Buttons = json["buttons"];
        Description = json["description"];
    }

    void draw() {
        vec2 imgSize = newsDisplayImage(Image);

        UI::BeginChild("news-image-buttons", imgSize);

        if (Buttons.GetType() == Json::Type::Array) {
            UI::PushFont(BigFont);
            for (uint i = 0; i < Buttons.Length; i++) {
                Json::Value json = Buttons[i];
                vec4 color = vec4(json["color"][0], json["color"][1], json["color"][2], json["color"][3]);
                UI::PushStyleColor(UI::Col::Button, color);
                UI::PushStyleColor(UI::Col::ButtonHovered, color*1.1);
                UI::PushStyleColor(UI::Col::ButtonActive, color*0.9);
                if (UI::Button(string(json["name"]) + " " + icons(json["icon"]))) {
                    if (json["type"] == "website") {
                        OpenBrowserURL(json["websiteUrl"]);
                    }
                    else if (json["type"] == "redirect") {
                        selectedTab = menuTabFromString(json["tab"]);
                        if (json.HasKey("filter_map")) {
                            mapsFilter.name = json["filter_map"];
                            mapsFilter.nameOn = true;
                        }
                        else if (json.HasKey("filterCampaing")) {
                            memory.Get(orderedKeys[json["filterCampaing"]]).playCampaign();
                        }
                    }
                }
                UI::PopStyleColor(3);
            }
            UI::PopFont();
        }

        UI::EndChild();

        if (Description.GetType() == Json::Type::Array) {
            for (uint i = 0; i < Description.Length; i++) {
                string type = Description[i]["type"];
                if (type == "text") drawDescriptionText(Description[i]);
            }
        }
    }

    void drawDescriptionText(Json::Value json) {
        if (json["type"] != "text") return;
        UI::Text(json["text"]);
    }

    vec2 newsDisplayImage(UI::Texture@ image, imgScaleType scaleType = imgScaleType::scaleToWidth) {
        UI::DrawList@ DL = UI::GetWindowDrawList();
        vec2 windowPos = UI::GetWindowPos();
        float offset = 90 - UI::GetScrollY();
        windowPos.y += offset;
        vec2 img = scaleImage(image.GetSize(), offset, scaleType);
        windowPos.x += UI::GetWindowSize().x/2.0f - img.x/2.0f;
        DL.AddImage(image, windowPos, img);
        return img;
    }
}