class tag {
    // save in memory
    string name = "";
    uint16[] rgbColor = {0, 0, 0}; // ranging from 0 - 255, call rgb(r, g, b) from main to get real color
    uint64 tagUid = 0; // 0 is reserved for no value
    tag::campaignMapId@[] maps = {};

    // get in startup
    uint mapsWithTag = 0;
    vec4 colorVec4 = vec4(0, 0, 0, 0);

    tag() {}

    tag(const string &in _name, uint16[] _rgbColor, uint64 _tagUid, tag::campaignMapId@[] _maps = {}) {
        name = _name;
        rgbColor = _rgbColor;
        tagUid = _tagUid;
        maps = _maps;

        colorVec4 = rgb(rgbColor[0], rgbColor[1], rgbColor[2]);
    }

    void setTagAtMaps() {
        for (uint i = 0; i < maps.Length; i++) {
            debug("settting tag at a map");
            startnew(CoroutineFuncUserdataUint64(maps[i].setTagAtMap), tagUid);
            if (i % 250 == 0) yield(); // give space
        }
    }

    void writeToBuffer(MemoryBuffer@ buf) {
        stringToBuffer(buf, name);
        uint16ArrayToBuffer(buf, rgbColor);
        buf.Write(tagUid);
        mapsWriteToBuffer(buf);
    }

    void mapsWriteToBuffer(MemoryBuffer@ buf) {
        uint16 len = maps.Length;
        buf.Write(len);
        for (uint i = 0; i < len; i++) {
            maps[i].writeToBuffer(buf);
        }
    }


    // the ammount of maps will never be bigger than a few thousend so using linear time is fine
    void addMap(mapInfo@ mi) { 
        mapsWithTag += 1;
        maps.InsertLast(tag::campaignMapId(mi.campaignId, mi.trackId));
    }

    void removeMap(mapInfo@ mi) {
        mapsWithTag -= 1;
        maps.RemoveAt(maps.Find(tag::campaignMapId(mi.campaignId, mi.trackId)));
    }

    void drawTableItem(bool map = false, mapInfo@ mi = mapInfo()) {
        UI::PushID(tagUid);
        UI::TableNextRow();
        UI::TableNextColumn();
        if (develepor) {
            UI::Text(""+tagUid);
            UI::TableNextColumn();
        }
        UI::PushStyleColor(UI::Col::Text, colorVec4);
        UI::Text(Icons::Circle);
        UI::PopStyleColor();
        UI::TableNextColumn();
        UI::Text(name);
        UI::TableNextColumn();
        UI::Text(""+mapsWithTag);
        UI::TableNextColumn();
        if (!map) { // display in tag menu
            UI::PushStyleColor(UI::Col::Button, ColorYellow);
            UI::PushStyleColor(UI::Col::ButtonHovered, ColorYellow*1.1);
            UI::PushStyleColor(UI::Col::ButtonActive, ColorYellow*0.9);
            if (UI::Button("Edit")) {
                startnew(CoroutineFunc(edit));
            }
            UI::PopStyleColor(3);
            UI::SameLine();

            UI::PushStyleColor(UI::Col::Button, ColorRed);
            UI::PushStyleColor(UI::Col::ButtonHovered, ColorRed*1.1);
            UI::PushStyleColor(UI::Col::ButtonActive, ColorRed*0.9);
            if (UI::Button("Delete")) {
                tryDelete();
            }
            UI::PopStyleColor(3);
        } else { // display in map info
            if (mi.hasTag(tagUid)) {
                UI::PushStyleColor(UI::Col::Button, ColorRed);
                UI::PushStyleColor(UI::Col::ButtonHovered, ColorRed*1.1);
                UI::PushStyleColor(UI::Col::ButtonActive, ColorRed*0.9);
                if (UI::Button(Icons::Times)) {
                    mi.removeTag(tagUid);
                }
                UI::PopStyleColor(3);
            } else {
                UI::PushStyleColor(UI::Col::Button, ColorGreen);
                UI::PushStyleColor(UI::Col::ButtonHovered, ColorGreen*1.1);
                UI::PushStyleColor(UI::Col::ButtonActive, ColorGreen*0.9);
                if (UI::Button(Icons::Plus)) {
                    mi.addTag(tagUid);
                }
                UI::PopStyleColor(3);
            }
        }
        UI::PopID();
    }

    void tryDelete() {
        debug("Adding confirmation window for tag deletion");
        string text = "Do you wanna delete the tag " + name + "(" + tagUid + ")?";
        confirmationWindow@ cw = confirmationWindow(text, CoroutineFunc(delete));
        confirmationWindows.InsertLast(cw);
    }

    void delete() {
        tag::deleteTag(tagUid);
    }

    void drawTagButton() {
        UI::PushStyleColor(UI::Col::Button, colorVec4);
        UI::PushStyleColor(UI::Col::ButtonHovered, colorVec4*1.1);
        UI::PushStyleColor(UI::Col::ButtonActive, colorVec4*0.9);
        if (UI::Button(name)) {
            startnew(CoroutineFunc(edit));
        }
        UI::PopStyleColor(3);
    }

    void edit() {
        editName = name;
        editColor = vec3(rgbColor[0]/255.,rgbColor[1]/255.,rgbColor[2]/255.);
        @tag::currentTagEdit = this;
        tag::showTagEdit = true;
        selectedTab = menuTabs::tagInfo;
    }

    void drawEditTabInner() {
        UI::BeginChild("editTab");

        UI::PushFont(BigFont);
        string title = name;
        if (develepor) title += "\\$444 (" + tagUid + ")\\$z";
        UI::Text(title);
        UI::PopFont();

        drawEdit();

        UI::Separator();

        drawMapTable();

        UI::EndChild();
    }

    string editName = "";
    vec3 editColor = vec3(0, 0, 0);
    void drawEdit() {
        editName = UI::InputText("Name", editName);
        editColor = UI::InputColor3("Color", editColor);

        vec4 editColor4 = vec3to4(editColor);

        UI::PushStyleColor(UI::Col::Button, editColor4);
        UI::PushStyleColor(UI::Col::ButtonHovered, editColor4*1.1);
        UI::PushStyleColor(UI::Col::ButtonActive, editColor4*0.9);
        
        UI::BeginDisabled(editName == "");

        if (UI::Button(Icons::Check + " Edit tag")) {
            name = editName;
            rgbColor = irgb(editColor);
            colorVec4 = editColor4;
            tagMemory.Set("" + tagUid, this);
        }

        UI::EndDisabled();

        UI::PopStyleColor(3);
    }

    void drawMapTable() {
        // setup map filter with this tag in inclusive
        mapFilter@ tagMapsFilter = mapFilter();
        tagMapsFilter.tagInclusive = {tagUid};
        mapInfo@[] mapsToShow = {};
        filterMaps(mapsToShow, tagMapsFilter);
        drawFilteredMapsTable(mapsToShow);
    }
}

namespace tag {
    uint64 nextTagUid = 1;
    tag@ currentTagEdit = null;
    bool showTagEdit = false;

    tag@ readFromBuffer(MemoryBuffer@ buf) {
        string name = StringFromBuffer(buf);
        uint16[] rgbColor = uint16ArrayFromBuffer(buf);
        uint64 tagUid = uint64FromBuffer(buf);
        campaignMapId@[] maps = mapsReadFromBuffer(buf);
        return tag(name, rgbColor, tagUid, maps);
    }

    campaignMapId@[] mapsReadFromBuffer(MemoryBuffer@ buf) {
        uint16 len = uint16FromBuffer(buf);
        campaignMapId@[] maps = {};
        for (uint i = 0; i < len; i++) {
            //campaignMapId@ cmi = campaignMapId::readFromBuffer(buf);
            //maps[i] = cmi;
            maps.InsertLast(campaignMapId::readFromBuffer(buf));
        }
        return maps;
    }

    void deleteTag(uint uid) {
        allTagUids.RemoveAt(allTagUids.Find(uid));
        tagMemory.Overide();
    }

    class campaignMapId {
        string campaignId;
        string mapId;

        campaignMapId() {}

        campaignMapId(const string &in _campaignId, const string &in _mapId) {
            campaignId = _campaignId;
            mapId = _mapId;
        }

        bool opEquals(campaignMapId@ cmi) {
            return (campaignId == cmi.campaignId && mapId == cmi.mapId);
        }

        void writeToBuffer(MemoryBuffer@ buf) {
            stringToBuffer(buf, campaignId);
            stringToBuffer(buf, mapId);
        }

        void setTagAtMap(uint64 tagUid) {
            while (!memory.In(campaignId)) yield();
            memory.Get(campaignId).loadTagAtMap(tagUid, mapId);
        }
    }

    namespace campaignMapId {
        campaignMapId@ readFromBuffer(MemoryBuffer@ buf) {
            string campaignId = StringFromBuffer(buf);
            string mapId = StringFromBuffer(buf);
            return campaignMapId(campaignId, mapId);
        }
    }

    namespace GUI {
        void drawTab() {
            if (!tagMemory.initialized) {
                UI::Text("Loading tags");
                return;
            }

            drawTagCreator();
            UI::Separator();
            drawTagInfo();
        }
        
        string creatorName = "";
        vec3 creatorColor = vec3(Math::Rand(0., 1.), Math::Rand(0., 1.), Math::Rand(0., 1.)); // random colors before user sets them
        void drawTagCreator() {
            creatorName = UI::InputText("Name", creatorName);
            creatorColor = UI::InputColor3("Color", creatorColor);
            
            vec4 creatorColor4 = vec3to4(creatorColor);

            UI::PushStyleColor(UI::Col::Button, creatorColor4);
            UI::PushStyleColor(UI::Col::ButtonHovered, creatorColor4*1.1);
            UI::PushStyleColor(UI::Col::ButtonActive, creatorColor4*0.9);
            
            UI::BeginDisabled(creatorName == "");

            if (UI::Button(Icons::Plus + " Create new tag")) {
                createNewTag();
            }

            UI::EndDisabled();

            UI::PopStyleColor(3);
        }

        void createNewTag() {
            tag@ newTag = tag(creatorName, irgb(creatorColor), nextTagUid);
            tagMemory.Set(""+nextTagUid, newTag);
            allTagUids.InsertLast(nextTagUid);

            debug("nextTagUid : " + nextTagUid);

            // reset tag creator
            nextTagUid++;
            creatorName = "";
            creatorColor = vec3(Math::Rand(0., 1.), Math::Rand(0., 1.), Math::Rand(0., 1.));
        }

        void drawTagInfo(bool map = false, mapInfo@ mi = mapInfo()) {
            /*
            UI::BeginChild("buffer", vec2(UI::GetWindowSize().x*0.05, 0));
            UI::EndChild();
            UI::SameLine();
            UI::BeginChild("tagInfo", vec2(UI::GetWindowSize().x*0.8, 0)); */
            if (UI::BeginTable("tags", develepor ? 5 : 4, UI::TableFlags::SizingFixedFit | UI::TableFlags::RowBg)) {
                if (develepor) UI::TableSetupColumn("uid", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("111").x);
                UI::TableSetupColumn("", UI::TableColumnFlags::WidthFixed, Draw::MeasureString(Icons::Circle).x);
                UI::TableSetupColumn("name", UI::TableColumnFlags::WidthStretch);
                UI::TableSetupColumn("maps", UI::TableColumnFlags::WidthFixed, Draw::MeasureString("11111").x);
                UI::TableSetupColumn("buttons");
                UI::TableHeadersRow();

                drawTagInfoInner(map, mi);

                UI::EndTable();
            }
            //UI::EndChild();
        }

        void drawTagInfoInner(bool map = false, mapInfo@ mi = mapInfo()) {
            UI::ListClipper clip(allTagUids.Length);
            while (clip.Step()) {
                for (int i = clip.DisplayStart; i < clip.DisplayEnd; i++) {
                    tagMemory.Get(""+allTagUids[i]).drawTableItem(map, mi);
                }
            }
        }
    }
}