class tagmem {
    dictionary @d;
    string filePath = IO::FromStorageFolder("tagmem.db");
    bool initialized = false;
    uint16 version = 0;

    tagmem() {
        @d = dictionary();
    }

    void LoadMem() {
        if (IO::FileExists(filePath)) {
            IO::File f(filePath, IO::FileMode::Read);
            MemoryBuffer@ buf = MemoryBuffer(f.Read(f.Size()));
            f.Close();

            version = buf.AtEnd() ? 0 : buf.ReadUInt16();
            trace("tag memory newest version : " + version + " ; newest version " + tagmem::tagDataBaseVersion);
            if (version != tagmem::tagDataBaseVersion) { // do not update tagDataBaseVersion with destructive update
                version = tagmem::tagDataBaseVersion;
                setVersion(tagmem::tagDataBaseVersion);
                initialized = true;
                return;
            }
            while (!buf.AtEnd()) {
                auto kvp = tagmem::kvPair::readFromBuffer(buf);
                @d[kvp.key] = kvp.val;
                if (allTagUids.Find(kvp.val.tagUid) == -1) allTagUids.InsertLast(kvp.val.tagUid);
                tag::nextTagUid = Math::Max(tag::nextTagUid, kvp.val.tagUid + 1);
                yield(); // yield every tag to give space for other things
            }
        }
        else {
            version = tagmem::tagDataBaseVersion;
            setVersion(tagmem::tagDataBaseVersion);
        }
        for (uint i = 0; i < allTagUids.Length; i++) {
            Get("" + allTagUids[i]).setTagAtMaps();
        }
        initialized = true;
    }

    void setVersion(uint16 v) {
        MemoryBuffer@ buf = MemoryBuffer();
        buf.Write(v);
        IO::File f(filePath, IO::FileMode::Write);
        f.Write(buf);
        f.Close();
    }

    tag@ Get(const string &in key) {
        return cast<tag@>(d[key]);
    }

    bool In(const string &in key) {
        return d.Exists(key);
    }

    string[] GetKeys() {
        return d.GetKeys();
    }

    void Set(const string &in key, tag@ val) {
        @d[key] = val;
        tagmem::kvPair@ kvp = tagmem::kvPair(key, val);
        MemoryBuffer@ buf = MemoryBuffer();
        kvp.writeToBuffer(buf);
        buf.Seek(0, 0);
        IO::File f(filePath, IO::FileMode::Append);
        f.Write(buf);
        f.Close();
    }

    void Clear() {
        allTagUids = {};
        d.DeleteAll();
        setVersion(tagmem::tagDataBaseVersion);
    }

    void Overide() {
        setVersion(tagmem::tagDataBaseVersion);
        for (uint i = 0; i < allTagUids.Length; i++) {
            Set(""+allTagUids[i], Get(""+allTagUids[i]));
        }
    }
}

namespace tagmem {
    uint16 tagDataBaseVersion = 5;

    class kvPair {
        string key;
        tag@ val = tag();

        kvPair(const string &in _key, tag@ _val) {
            val = _val;
            key = _key;
        }

        void writeToBuffer(MemoryBuffer@ buf) {
            stringToBuffer(buf, key);
            val.writeToBuffer(buf);
        }
    }

    namespace kvPair {
        kvPair@ readFromBuffer(MemoryBuffer@ buf) {
            string key = StringFromBuffer(buf);
            tag@ val = tag::readFromBuffer(buf);
            return kvPair(key, val);
        }
    }
}